* {
    background-color: black;
    background-image: none;
    color: white;
    font-family: Verdana;
}

/* .slider and .trough are parts of scrollbar */
.slider, .progressbar {
    border: none;
    background-color: white;
}

/* The bottom text is automatically sized based on the resolution of
 * the presenter's screen but setting this value in the user's custom
 * CSS will override the auto-sizing.  See the man page for more
 * details.
    .bottomText {
        font-size: 5em;
    }
 */

GtkProgressBar {
    color: gray;
}

.trough {
    border: none;
}

pdfpcTimerLabel.pretalk {
    color: green;
}

pdfpcTimerLabel.last-minutes {
    color: orange;
}

pdfpcTimerLabel.overtime {
    color: red;
}

